#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "greedy.h"

void nj_Rij(float **d, int n);
void nj_Sij(int *ii, int *jj, float **d, int *t, float *la);
void nj_reduction(int ii, int jj, float **d);
void nj_sommets(float **d, int *t, float *la);
int nj(float **d, int nsm, int *t, float *la);

int sk(float **d, int nsm, int *pere, float *la, int groupe_ext);

void nj_afficher(int nlig, int ncol);
void nj_ecrire(int x, int y, char car);
void nj_effacer(int nlig, int ncol);
void nj_ecrire_chaine(int x, int y, char *chaine);
void nj_dessin_branches(int x, int y1, int y2);
void nj_dessin_arbre(noeud *base, int niveau, int *position);
noeud *nj_calculer_arbre(score **d);


extern score DISTANCE_MAX;
extern gscore SCORE_MIN;
extern int NBRE_LETTRES, X, INCONNU;
extern sequence *seq;
extern indice maxlong, minlong;
extern int Ra, R, nseqbasemin, nseqmin;
extern long RL;
extern indice longalig;

extern score **sim, max_sim;

extern indice nech;

extern char plein, vide;
extern int long_branche, separation;
extern int num_feuille;
extern char **grille;
extern noeud **noeuds;

extern char DEBUG;

int n; 
int ns, nf, *id;


void nj_Rij(float **d, int n)
{
/*place the sum of the distances between i and all others in d[i][i].
*d[] must be symmetrized before*/

  int i,j; float smd;

  smd=0;
  for(j=1;j<=n;j++){ 
    		smd=0;
    		for(i=1;i<=n;i++) if(i!=j) smd+=d[i][j];
    		d[j][j]=smd;
  		}     
}

void nj_Sij(int *ii, int *jj, float **d, int *t, float *la)
{
  int i,j; float sij,smin;
 
  smin=0;

  for(i=1;i<=n;i++)
    for(j=i+1;j<=n;j++){
  
      	sij=d[i][j]*(n-2)-d[i][i]-d[j][j];

	if(sij<=smin) {*ii=i; *jj=j; smin=sij;}   /*search for the smallest sij*/
    	}
  
  ns+=1; t[id[*ii]]=ns; t[id[*jj]]=ns;

 
  la[id[*ii]]=((n-2)*d[*ii][*jj]+d[*ii][*ii]-d[*jj][*jj])/(2*(n-2));
  la[id[*jj]]=((n-2)*d[*ii][*jj]-d[*ii][*ii]+d[*jj][*jj])/(2*(n-2));
}

void nj_reduction(int ii,int jj, float **d)
{
  int i;

  for(i=1;i<=n;i++) if(i!=ii && i!=jj){
     			d[i][ii]=(d[i][ii]+d[i][jj]-d[ii][jj])/2; d[ii][i]=d[i][ii];
   			 }
  
  d[ii][ii]=0; id[ii]=ns; 

  for(i=1;i<n;i++){				/* n->jj */
    	d[i][jj]=d[i][n]; d[jj][i]=d[n][i];
  	}

  d[jj][jj]=0; id[jj]=id[n];
       
  n-=1;  
}

void nj_sommets(float **d, int *t, float *la)
{
  int i,j,k;

  i=id[1]; j=id[2]; k=id[3];
  ns+=1; t[i]=t[j]=t[k]=ns;
  
  la[i]=(d[1][2]+d[1][3]-d[2][3])/2;
  la[j]=(d[1][2]+d[2][3]-d[1][3])/2;
  la[k]=(d[1][3]+d[2][3]-d[1][2])/2;
}



int nj(float **d, int nsm, int *t, float *la)
/*algorithme de Studier-Keppler*/
/* recoit une matrice de distance d entre nsm sequences,
retourne :  	- le nombre de noeuds de l'arbre 
		- dans t: pour chaque noeud l'indice du noeud pere
		- dans la: pour chaque noeud la distance au noeud pere 
les indices des sequences vont de 1 a nsm dans d, t, et la
les indices des noeuds vont de 1 a ns-1 (ns=racine) pour t et la */
{
int it,i,ii,jj;

/*memory allocations for arrays*/
/*for array *id, subscript numbering goes from 1 to nsm, not from 0 to nsm-1. Thus, nsm+1 elements are allocated in order to compute id[nsm]*/

id = (int *) allouer((nsm+1) * sizeof(int));

ns=nf=n=nsm; 
for(i=1;i<=n;i++) id[i]=i;

it=0;
while (n>3){
    		it+=1;
		nj_Rij(d,n);
    		nj_Sij(&ii,&jj, d, t, la);
    		nj_reduction(ii,jj, d);  
  		}
nj_sommets(d, t, la);

 
for(i=1;i<=ns;i++) if(la[i]<=0.0) la[i]=0.0;  /*NO NEGATIVE BRANCH LENGTHS!!*/

liberer(id);

/*for (i=1; i < ns; i++)
printf("noeud %d fils de %d (%f)\n", i, t[i], la[i]);*/

return(ns);
}


void nj_afficher(int nlig, int ncol)
{
int x, y;

for (y=0; y < nlig; y++, printf("\n"))
    for (x=0; x < ncol; x++)
   	printf("%c", grille[y][x]);
printf("\n");
}

void nj_ecrire(int x, int y, char car)
{

grille[y][x] = car;
}

void nj_effacer(int nlig, int ncol)
{
int x, y;

for (y=0; y < nlig; y++)
    for (x=0; x < ncol; x++)
	ecrire(x, y, vide);
}

void nj_ecrire_chaine(int x, int y, char *chaine)
{
int i;

for (i=0; chaine[i] != 0; i++)
	ecrire(x+i, y, chaine[i]);
}

void nj_dessin_branches(int x, int y1, int y2)
{

int i;

for (i=y1; i <= y2; i++)
	ecrire(x, i, plein);

for (i=1; i <= long_branche; i++)
    {
	ecrire(x+i, y1, plein);
	ecrire(x+i, y2, plein);
    }
}

void nj_dessin_arbre(noeud *base, int niveau, int *position)
{
int position1, position2;
char num[10];

if (base->noeud1 == NULL)
   {
	*position = num_feuille * separation;

	sprintf(num, "%d", base->num);

	ecrire_chaine(niveau+1, *position, num);

	num_feuille++;
   }
else
   {
	dessin_arbre(base->noeud1, niveau+long_branche+1, &position1);

	dessin_arbre(base->noeud2, niveau+long_branche+1, &position2);

	*position = (position1 + position2) / 2;

	dessin_branches(niveau, position1, position2);
   }
}

int sk(float **d, int nsm, int *pere, float *la, int groupe_ext)
/* recoit une matrice de distance d entre nsm sequences,
retourne :  	- le nombre de noeuds de l'arbre 
		- dans pere: pour chaque noeud l'indice du noeud pere
		- dans la: pour chaque noeud la distance au noeud pere 
les indices des sequences vont de 1 a nsm dans d, pere, et la
les indices des noeuds vont de 1 a la valeur retournee pour pere et la 
ATTENTION: la matrice d est affect par sk */
{
int taxon, taxon_suiv, taxon_suiv_suiv, nbretaxons;
double lg_branche_suiv, lg_branche;

nbretaxons = nj(d, nsm, pere, la);

taxon = pere[groupe_ext]; 
taxon_suiv = pere[taxon];
lg_branche = la[taxon];

while (taxon != nbretaxons)
   {
	taxon_suiv_suiv = pere[taxon_suiv];
	lg_branche_suiv = la[taxon_suiv];

	pere[taxon_suiv] = taxon;
	la[taxon_suiv] = lg_branche;

	taxon = taxon_suiv;
	taxon_suiv = taxon_suiv_suiv;
	lg_branche = lg_branche_suiv;
   }

taxon=pere[groupe_ext]; 
pere[taxon] = 0;
la[taxon] = 1.;

return nbretaxons;
}

noeud *nj_calculer_arbre(score **d)
{
int x, y, z, s, k, nbretaxons, groupe_ext, racine;
score *nbre, dist, dist_min;
noeud  *nnoeud, **tnoeuds;
char *supprimee;
float **dd, *la; 
int *pere;

DISTANCE_MAX = - (SCORE_MIN.b + 1);

if (DEBUG)	printf("nj_calculer_arbre\n");

pere = (int *) allouer(2 * R * sizeof(int));
la = (float *) allouer(2 * R * sizeof(float));
dd = (float **) callouer_mat(sizeof(float), R+1, R+1);
tnoeuds = (noeud **) allouer(R * sizeof(noeud *));

dist = groupe_ext = 0;

for (x=0; x < R; x++)
   {
	noeuds[x] = (noeud *) allouer(sizeof(noeud));
printf("%d ", noeuds[x]);
	noeuds[x]->noeud1 = noeuds[x]->noeud2 = NULL;
	noeuds[x]->num = x;

	noeuds[x]->premier = (int *) allouer(sizeof(int));
	noeuds[x]->dernier = (int *) allouer(sizeof(int));
	noeuds[x]->premier[0] = 1;
	noeuds[x]->dernier[0] = noeuds[x]->longueur = seq[x].longueur;

	for (y=dist=0; y < R; y++)
	   {
		dd[x+1][y+1] = (float) d[x][y];
		dist += d[x][y];
	   }

	if (dist > dist_min)
	   {
		dist = dist_min;
		groupe_ext = x;
	   }
   }

if (DEBUG)	printf("groupe externe : %d\n", groupe_ext);

nbretaxons =  sk(dd, R, pere, la, groupe_ext+1);

for (x=0; x < nbretaxons; x++)
   {
	tnoeuds[x] = (noeud *) allouer(sizeof(noeud));
	pere[x] = pere[x+1] - 1;
	la[x] = la[x+1];
   }

z = pere[groupe_ext];
racine = nbretaxons;
pere[groupe_ext] = pere[z] = racine;
pere[racine] = 0;
tnoeuds[racine-R] = (noeud *) allouer(sizeof(noeud));
nbretaxons++;

if (DEBUG)
   {	printf("nbre de taxons = %d\n", nbretaxons);
	for (x=0; x < nbretaxons; x++)
		printf("%d est fils de %d (%d)\n", x, pere[x], (int)la[x]);
   }

for (k=R; k < nbretaxons; k++)
   {
	for (x=0; x < nbretaxons && pere[x] != k; x++);
	for (y=x+1; y < nbretaxons && pere[y] != k; y++);

	if (x < R) 	tnoeuds[k-R]->noeud1 = noeuds[x]; 
	else		tnoeuds[k-R]->noeud1 = tnoeuds[x-R];
	if (y < R) 	tnoeuds[k-R]->noeud2 = noeuds[y]; 
	else		tnoeuds[k-R]->noeud2 = tnoeuds[y-R];
	tnoeuds[k-R]->num = k;
   }

for (k=0; k < R; k++)
printf("%d=%d: %d\n", k, noeuds[k]->num, noeuds[k]);
for (k=R; k < nbretaxons; k++)
printf("%d=%d: %d est pere de %d (%d) et de %d (%d)\n", k, tnoeuds[k-R]->num, tnoeuds[k-R], tnoeuds[k-R]->noeud1,tnoeuds[k-R]->noeud1->num, tnoeuds[k-R]->noeud2, tnoeuds[k-R]->noeud2->num);

nnoeud = tnoeuds[racine-R];


liberer(pere);
liberer(la);
liberer(tnoeuds);
liberer_mat((void **) dd, R+1);

if (DEBUG)	printf("nj_calculer_arbre termine'm retourne %d\n", nnoeud);

return(nnoeud);
}

